# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Snappy Ecosystem Tests
# Copyright (C) 2017 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Manage tests config files: get existing one, create new one and store it"""

import configparser
import logging
import os

LOGGER = logging.getLogger(__name__)

DEFAULT_CONF = 'ecosystem_tests.cfg'
KEY_DEFAULT = 'default'


def get_config_dir():
    """Return the path to the tests configuration directory."""
    conf_dir = os.environ.get(
        'XDG_CONFIG_HOME', os.path.realpath(__file__ + '/../../configs'))
    return conf_dir


def get_user_config_dir():
    """Return the path to the user configuration directory."""
    conf_dir = os.environ.get(
        'XDG_USER_CONFIG_HOME', os.path.expanduser('~/.config'))
    return conf_dir


def get_config_section():
    """Return the name of config section to use."""
    return os.environ.get('CONFIG_SECTION', KEY_DEFAULT)


def _get_config_stack_from_file(file_path):
    """Return config stack from specified file path."""
    ecosystem_config = SnappyEcosystemTestsConfig(file_path)
    ecosystem_config.read_config_from_file()
    return ecosystem_config


class ConfigStack:
    """Class that represents different config stacks"""

    def __init__(self, dir_path=None, file_name=None):
        self.__config = self.__create_config_stack(dir_path, file_name)

    @property
    def config(self):
        """Return the current config"""
        return self.__get_config()

    def set_config(self, dir_path=None, file_name=None):
        """Set the current config for the given dirpath and file name.
        :param dir_path: the dir in which the config file is located.
        If None it takes the default one returned by get_config_dir()
        :param file_name: the config file name.
        If None it takes the default one stored in DEFAULT_CONF
        :return the created config as a dict
        """
        self.__config = self.__create_config_stack(dir_path, file_name)

    def release_config(self):
        """Release the current config for those cases that
         explicitly want to modify it"""
        self.__config = None

    def __get_config(self):
        """Get current config. Set it in the case is None"""
        if self.__config is None:
            self.set_config()
        return self.__config

    @staticmethod
    def __create_config_stack(dir_path=None, file_name=None):
        """Create the config stack for the given dirpath and file name
        :param dir_path: the dir in which the config file is located.
        If None it takes the default one returned by get_config_dir()
        :param file_name: the config file name.
        If None it takes the default one stored in DEFAULT_CONF
        :return the created config as a dict
        """
        full_path = os.path.join(dir_path or get_config_dir(),
                                 file_name or DEFAULT_CONF)
        return _get_config_stack_from_file(full_path)


class SnappyEcosystemTestsConfig:

    """Class to get and save test configuration data."""

    def __init__(self, file_name=None):
        """
        Construct the config and parser object.
        :param file_name: Name of config file to load. Default file is used if
        value is not specified.
        """
        user_config_dir = get_config_dir()
        if not os.path.exists(user_config_dir):
            os.mkdir(user_config_dir)
        self.file_path = os.path.join(
            user_config_dir, file_name or DEFAULT_CONF)
        self.config = configparser.ConfigParser(
            allow_no_value=True, empty_lines_in_values=False)

    def set(self, section, option_name, value):
        """Set a config value for the given section.
        :param section: the section of the config in which update
         the given value
        :param option_name: Name of config option.
        :param value: Value to set.
        """
        self.config[section][option_name] = value
        self.save()

    def get(self, section, option_name, default=''):
        """Get a config value for a given section.
        If section and/or value are not found, return default if
        provided or raise exception
        :param section: the section of the config in which search
        for the given value.
        :param option_name: Name of config option to get.
        :param default: Value to return as default if section and/or
        item don't exist.
        :return: Config option value if it exists. In case that section
        and/or item don't exists,
        then return default value if provided, or raise exception"""
        try:
            return self.config.get(section, option_name)
        except(configparser.NoSectionError, configparser.NoOptionError):
            if default:
                return default
            else:
                raise

    def get_section(self, section):
        """
        Return a particular section of config
        :param section: the section name to retrieve
        :return: a Section object
        """
        return self.config[section]

    def save(self):
        """Save the config to file path"""
        mode = 'r+' if os.path.exists(self.file_path) else 'w'
        with open(self.file_path, mode) as _f:
            self.config.write(_f)

    def read_config_from_file(self):
        """Read config from file"""
        self.config.read(self.file_path)


CONFIG_STACK = ConfigStack().config
USER_CONFIG_STACK = ConfigStack(dir_path=get_user_config_dir()).config
