# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2017 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Managers to setup the environment"""

import os
import shlex
from pylxd import Client
from pylxd.exceptions import NotFound, LXDAPIException
from retrying import retry

from snappy_ecosystem_tests.environment import constants
from snappy_ecosystem_tests.environment.constants import (
    PROFILES, DEPENDENCIES)
from snappy_ecosystem_tests.environment.data.snapcraft import (
    SNAPCRAFT_CONTAINER_NAME)
from snappy_ecosystem_tests.environment.data.snapd import (
    DEFAULT_CONTAINER_CONFIG, SNAPD_CONTAINER_NAME)

COMMAND_ALLOW_ENV_VARS = 'sed -i \'/^AcceptEnv/ s/$/ {}/\' ' \
                         '/etc/ssh/sshd_config'


def get_manager(mode):
    """Get the manager for the given mode"""
    supported_modules = {
        "lxd": LxdManager
    }
    try:
        manager = supported_modules[mode]
        return manager()
    except KeyError:
        raise RuntimeError("Mode: {} is not supported".format(mode))


class LxdManager:
    """Manage lxd containers"""

    def __init__(self):
        self.client = Client()

    def launch(self, container_name, release='16.04'):
        """Launch a container"""
        try:
            container = self.client.containers.get(container_name)
            if container.status == "Running":
                container.stop(wait=True)
            container.delete(wait=True)
        except NotFound:
            pass
        container_config = DEFAULT_CONTAINER_CONFIG.copy()
        container_config['source']['alias'] = release
        container_config['name'] = container_name
        container = self.client.containers.create(container_config, wait=True)
        container.start(wait=True)

    def execute_command(self, container_name, command):
        """execute command"""
        container = self.client.containers.get(container_name)
        container.execute(command)

    def install(self, container_name, packages, channel=None):
        """Install a packages"""
        container = self.client.containers.get(container_name)
        container.execute(shlex.split('apt-get update'))
        install_command = 'apt-get install {} -y'.format(
            ' '.join(packages))
        if channel:
            install_command += ' -t {}'.format(channel)
        container.execute(shlex.split(install_command))

    def install_dependencies(self, container_name, container_type):
        """Install a dependencies"""
        self.install(
            container_name,
            DEPENDENCIES[container_type] + DEPENDENCIES['shared'])

    def configure(self, container_name, container_type, profile):
        """Configure a given container"""
        container = self.client.containers.get(container_name)
        for key, value in \
                PROFILES[profile][container_type]['environment_variables'].\
                        items():
            container.config.update({'environment.{}'.format(
                key): value})
        container.save(wait=True)

    @retry(stop_max_attempt_number=5, wait_fixed=3000,
           retry_on_exception=lambda exception: isinstance(
               exception, LXDAPIException))
    def enable_ssh(self, container_name, container_type, profile):
        """Enable the ssh connection on the container"""
        container = self.client.containers.get(container_name)
        pub_key = open(
            os.path.expanduser('~') + '/.ssh/id_rsa.pub').read().strip('\n')
        container.files.put('/home/ubuntu/.ssh/authorized_keys', pub_key)
        # Allow export of environment variables over ssh.
        container.execute(shlex.split(COMMAND_ALLOW_ENV_VARS.format(
            ' '.join(PROFILES[profile][container_type]
                     ['environment_variables'].keys()))
        ))
        container.execute(shlex.split('service ssh restart'))

    def setup(self, profile):
        """setup the container based on the profile"""
        self._setup(SNAPD_CONTAINER_NAME, constants.SNAPD, profile)
        self._setup(
            SNAPCRAFT_CONTAINER_NAME, constants.SNAPCRAFT, profile)

    def _setup(self, container_name, container_type, profile):
        """
        Launch a container, enable ssh, install dependencies
        and setup the needed environment variables
        """
        self.launch(container_name)
        self.enable_ssh(container_name, container_type, profile)
        self.install(
            container_name,
            packages=[PROFILES[profile][container_type]["package_name"]],
            channel=PROFILES[profile][container_type]["channel"])
        self.install_dependencies(container_name, container_type)
        self.configure(container_name, container_type, profile)

    def get_ip(self, container_name):
        """Gets the ip address for a given container"""
        networks = self.client.containers.get(
            container_name).state().network['eth0']['addresses']

        for network in networks:
            if network['address']:
                return network['address']
        raise RuntimeError(
            "The container {} does not have an IPV4 connection".format(
                container_name))
