# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Snappy Ecosystem Tests
# Copyright (C) 2017 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Module to build a snap in a remote machine."""

import os
from pathlib import PurePath

from snappy_ecosystem_tests.utils.ssh import SSHManager

COMMANDS_SETUP = [
    'apt update',
    'apt dist-upgrade -y',
    'apt install snapcraft -y'
]
SNAPCRAFT_YAML = """\
name: {snap_name}
version: {version}
summary: {summary}
description: |
  {description}

grade: {grade}
confinement: {confinement}

parts:
  {snap_name}:
    plugin: nil

apps:
  {snap_name}:
    command: echo ok
"""


class SnapBuilder:
    """Class to build a snap on a remote machine and download its content."""
    def __init__(self, hostname, username, port=22):
        self.client = SSHManager.get_instance(hostname, username, port)

    def is_host_setup(self):
        """Return bool representing whether the remote host is setup by
        checking if snapcraft package is installed."""
        try:
            self.client.exec_command('which snapcraft')
            return True
        except ValueError:
            return False

    def setup_host(self):
        """Setup the host to build a snap, currently only installs
        snapcraft."""
        for command in COMMANDS_SETUP:
            self.client.exec_command(command)

    def build(self, name, version, summary, description, grade='stable',
              confinement='strict'):
        """Build a new snap in the remote host and return its remote path.

        :param name: name of the snap to build.
        :param version: version number for the snap to build.
        :param summary: short summary of the snap to build.
        :param description: long description of the snap.
        :param grade: stability grade of the snap e.g. stable or devel.
        :param confinement: confinement policy for the snap e.g.
            classic, devmode or strict.
        :raises ValueError: If an error is occurred during snap build.
        :return: Remote path of the newly build snap package.
        """
        if not self.is_host_setup():
            self.setup_host()
        tempdir = self.client.exec_command('mktemp -d')
        self.client.exec_command('mkdir snap', cwd=tempdir)
        self.client.exec_command('echo "{}" > snap/snapcraft.yaml'.format(
            SNAPCRAFT_YAML.format(
                snap_name=name,
                version=version,
                summary=summary,
                description=description,
                grade=grade,
                confinement=confinement
            )
        ), cwd=tempdir)
        self.client.exec_command('snapcraft', cwd=tempdir)
        return os.path.join(
            tempdir,
            '{name}_{version}_{arch}.snap'.format(
                name=name,
                version=version,
                arch=self.client.exec_command('dpkg --print-architecture')
            )
        )

    def build_and_pull(self, name, version, summary, description,
                       grade='stable', confinement='strict',
                       output_location='.'):
        """Build a snap based on the provided arguments and download it.

        :param name: name of the snap to build.
        :param version: version number for the snap to build.
        :param summary: short summary of the snap to build.
        :param description: long description of the snap.
        :param grade: stability grade of the snap e.g. stable or devel.
        :param confinement: confinement policy for the snap e.g.
            classic, devmode or strict.
        :param output_location: Local location to download the newly built
            snap to. Defaults to pwd.
        :raises ValueError: If an error is occurred during snap build.
        :return: Local path of the newly built and downloaded snap package.
        """
        remote_snap_path = self.build(
            name=name,
            version=version,
            summary=summary,
            description=description,
            grade=grade,
            confinement=confinement
        )
        output_path_absolute = os.path.abspath(
            os.path.join(output_location, PurePath(remote_snap_path).name)
        )
        self.client.pull(remote_snap_path, output_path_absolute)
        return output_path_absolute
