# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Snappy Ecosystem Tests
# Copyright (C) 2017 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Snapcraft client helpers"""

import logging
import subprocess
import time

from snappy_ecosystem_tests.models.snap import Snap
from snappy_ecosystem_tests.utils import ssh
from snappy_ecosystem_tests.utils.commands import build_command
from snappy_ecosystem_tests.utils.filters import filter_list
from snappy_ecosystem_tests.utils.storeconfig import get_current_store

from snappy_ecosystem_tests.commons.config import CONFIG_STACK

from snappy_ecosystem_tests.utils import storeconfig

# login credentials gotten from config file or shell environment
from snappy_ecosystem_tests.utils.user import (
    get_snapcraft_remote_host_credentials
)

LOGIN_EMAIL, LOGIN_PASSWORD = storeconfig.get_store_credentials()

# Commands
PATH_SNAPCRAFT = '/usr/bin/snapcraft'
COMMANDS_LOGIN = """\
/usr/bin/expect \
-c 'spawn {snapcraft} login' \
-c 'expect Email:' \
-c 'send {email}\\r' \
-c 'expect Password:' \
-c 'send {password}\\r' \
-c 'interact'\
"""

COMMAND_LOGOUT = 'logout'

COMMAND_REGISTER = 'register'
COMMAND_LIST_REGISTERED = 'list-registered'

LOGGER = logging.getLogger(__name__)

HOSTNAME, USERNAME, PORT = get_snapcraft_remote_host_credentials()
SNAPCRAFT_CONFIG_FILE_PATH = '~/.config/snapcraft/snapcraft.cfg'


class Snapcraft(object):
    """Contain Snapcraft specific functionality to use via command
    line interface"""
    def __init__(self):
        """Create new snapcraft instance."""
        self._login = False
        self._cleanup()

    @staticmethod
    def _run_snapcraft_command_ssh(command, debug=True):
        """
        Run snapcraft command via ssh
        :param command: the command to be executed.
        Can be a single string or a list of strings
        :param debug: whether to enable debug mode
        :return: the command's output
        """
        if not command.startswith(PATH_SNAPCRAFT):
            command = build_snapcraft_command(command)
        if debug:
            command += ' -d'
        return ssh.run_command(command, hostname=HOSTNAME, username=USERNAME,
                               port=PORT)

    @staticmethod
    def assert_logged_in():
        """Assert that an user is logged.
        :raise ValueError: in case login is false
        """
        if Snapcraft.is_logged_in() is False:
            raise ValueError("User is not logged in, "
                             "please login before using this command")

    @staticmethod
    def is_logged_in():
        """Return bool representing if the user is logged into snapcraft."""
        try:
            output = ssh.run_command(
                'cat %s' % SNAPCRAFT_CONFIG_FILE_PATH,
                hostname=HOSTNAME, username=USERNAME, port=PORT)
            return output is not ''
        except ValueError:
            return False

    def _cleanup(self):
        """Perform cleanup actions"""
        self.logout()

    @staticmethod
    def logout():
        """logout of snapcraft store session
        :return: True if logout was successful, False otherwise
        """
        Snapcraft._run_snapcraft_command_ssh(COMMAND_LOGOUT)
        return not Snapcraft.is_logged_in()

    @staticmethod
    def login(email=LOGIN_EMAIL, password=LOGIN_PASSWORD):
        """Login to Snapcraft. If email and/or password are not provided,
        use the default ones.
        :param email: the SSO email
        :param password: the email password
        """
        ssh.run_command(
            COMMANDS_LOGIN.format(email=email, password=password,
                                  snapcraft=PATH_SNAPCRAFT),
            hostname=HOSTNAME, username=USERNAME, port=PORT)
        return Snapcraft.is_logged_in()

    def list_registered(self, debug=True, raw=False):
        """Call snapcraft list-registered command,
        raise exception if not logged in
        :param debug: whether to enable debug mode
        :param raw: whether to return the output as a raw.
        If False, return a list of Snap instances
        """
        self.assert_logged_in()
        registered_snaps_output = self._run_snapcraft_command_ssh(
            COMMAND_LIST_REGISTERED,
            debug=debug)
        if raw:
            return registered_snaps_output
        else:
            snaps = registered_snaps_output.splitlines()
            # if debug is set, the first 3 lines of output are the API calls
            return parse_snaps_raw(snaps if debug is False else snaps[3:])

    def filter_snaps(self, **predicates):
        """
        Return a list of the snaps that match with all the given predicates.
        Examples:
        filter_snap() --> return all the snaps
        filter_snap(name='my_name') --> return the snaps that matches with
         name 'my_name'
        filter_snap(name='my_name', private=True) --> return the snaps that
         matches with name 'my_name' and are private.
        filter_snap(price='5') --> return the snaps with price '5'.
        :return: A list of snap instances
        """
        return filter_list(self.list_registered(raw=False), **predicates)

    def register(self, snap_name, private=False, wait=True):
        """
        Register a snap name in the store.
        :param snap_name: the snap name to register.
        :param private: whether the snap name is private.
        :param wait: wheter to wait after registration to hitting the store
         restriction on following registrations
        :return: True if the snap was registered successfully, False otherwise.
        """
        self.assert_logged_in()
        LOGGER.info('About to register snap: %s', snap_name)
        command = build_snapcraft_command(COMMAND_REGISTER, snap_name)
        if private:
            command += ' --private'
        try:
            Snapcraft._run_snapcraft_command_ssh(command)
        except subprocess.CalledProcessError as _e:
            LOGGER.error('Unable to register snap: %s', _e.output)
            return False
        if wait:
            time.sleep(int(CONFIG_STACK.get('time_between_registrations',
                                            get_current_store())))
        LOGGER.info('Snap %s registered successfully!!!', snap_name)
        return True


def build_snapcraft_command(*args, base_command=PATH_SNAPCRAFT):
    """
    Build a snapcraft command with arguments
    :param args: list of arguments for the command
    :param base_command: snapcraft executable that is going to be used
    :return: the snapcraft command ready to be executed
    """
    return build_command(base_command, *args)


def parse_snaps_raw(raw):
    """
    Parse snaps raw output to a list of Snaps object
    :param raw: the raw that contains the snaps info.
    It assumes that the first line is the output Header like:
     (Name, Since, Price, etc.) to get the attributes order.
    :return: A list of snaps instances
    """
    attributes_order = {}
    for _i, att in enumerate(raw[0].split()):
        attributes_order[att.lower()] = _i
    snaps = []
    for snap_raw in raw[1:]:
        attrs = snap_raw.split()
        private = attrs[attributes_order.get('visibility')].lower() == 'private'
        snaps.append(Snap(name=attrs[attributes_order.get('name')],
                          since=attrs[attributes_order.get('since')],
                          private=private,
                          price=attrs[attributes_order.get('price')],
                          notes=attrs[attributes_order.get('notes')]))
    return snaps
