# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Snappy Ecosystem Tests
# Copyright (C) 2017 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Helpers to store/manage/load snapcraft configuration in a config file"""

import configparser
import logging
import os
import urllib.parse

from snappy_ecosystem_tests.commons.config import CONFIG_STACK
from xdg import BaseDirectory

LOCAL_CONFIG_FILENAME = '.snapcraft/snapcraft.cfg'

LOGGER = logging.getLogger(__name__)


class SnapcraftConfig(object):
    """Hold configuration options in sections.
    There can be two sections for the sso related credentials: production and
    staging. This is governed by the UBUNTU_SSO_API_ROOT_URL environment
    variable. Other sections are ignored but preserved.
    """

    def __init__(self):
        self.parser = configparser.ConfigParser()
        self.filename = None
        self.load()

    @staticmethod
    def _section_name():
        """Return a section"""
        # The only section we care about is the host from the SSO url
        url = os.environ.get('UBUNTU_SSO_API_ROOT_URL',
                             CONFIG_STACK.config.get('production_urls', 'sso'))
        return urllib.parse.urlparse(url).netloc

    def get(self, option_name):
        """Get an option name for a given section"""
        try:
            return self.parser.get(self._section_name(), option_name)
        except (configparser.NoSectionError,
                configparser.NoOptionError,
                KeyError):
            return None

    def set(self, option_name, value):
        """Set an option with the given value"""
        section_name = self._section_name()
        if not self.parser.has_section(section_name):
            self.parser.add_section(section_name)
        return self.parser.set(section_name, option_name, value)

    def is_empty(self):
        """Return True if snapcraft config does not have the default section
         name"""
        # Only check the current section
        section_name = self._section_name()
        if self.parser.has_section(section_name):
            if self.parser.options(section_name):
                return False
        return True

    def load(self):
        """Load the snapcraft configuration"""
        # Local configurations (per project) are supposed to be static.
        # That's why it's only checked for 'loading' and never written to.
        # Essentially, all authentication-related changes, like login/logout
        # or macaroon-refresh, will not be persisted for the next runs.
        if os.path.exists(LOCAL_CONFIG_FILENAME):
            self.parser.read(LOCAL_CONFIG_FILENAME)
            LOGGER.warning(
                'Using local configuration (`%s`), changes will '
                'not be persisted.', LOCAL_CONFIG_FILENAME)
            return

        self.filename = BaseDirectory.load_first_config(
            'snapcraft', 'snapcraft.cfg')
        if self.filename and os.path.exists(self.filename):
            self.parser.read(self.filename)

    @staticmethod
    def save_path():
        """Return the save path for the snapcraft configuration"""
        return os.path.join(BaseDirectory.save_config_path('snapcraft'),
                            'snapcraft.cfg')

    def save(self):
        """Save the current configuration in the save path"""
        self.filename = self.save_path()
        with open(self.filename, 'w') as _f:
            self.parser.write(_f)

    def clear(self):
        """Clear the current configuration"""
        self.parser.remove_section(self._section_name())
