# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Snappy Ecosystem Tests
# Copyright (C) 2017 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Helpers around snapd command-line interface."""

import json
import logging

import yaml

from snappy_ecosystem_tests.utils import ssh
from snappy_ecosystem_tests.utils.user import get_snapd_remote_host_credentials

PATH_SNAP = '/usr/bin/snap'
COMMAND_DOWNLOAD = 'download {snap} --channel={channel}'
COMMAND_FIND = 'find {search_term}'
COMMAND_INFO = 'info {snap}'
COMMAND_INSTALL = 'install {snap} --channel={channel}'
COMMAND_LIST = 'list'
COMMAND_LOGIN = 'login {email}'
COMMAND_LOGOUT = 'logout'
COMMAND_REFRESH = 'refresh {snap} --channel={channel}'
COMMAND_REMOVE = 'remove {snap}'
CHANNEL_STABLE = 'stable'
COMMANDS_LOGIN = """\
/usr/bin/expect \
-c 'spawn snap login {email}' \
-c 'expect \"Password*\"' \
-c 'send {password}\\r' \
-c 'interact'\
"""
LOGGER = logging.getLogger(__name__)

HOSTNAME, USERNAME, PORT = get_snapd_remote_host_credentials()


def run_snapd_command_ssh(parameters, cwd=''):
    """Run snapd command over ssh.

    :param parameters: a string containing parameters for the `snap` command.
    :param cwd: the current working directory on the remote where the command
        should run.
    :returns: stdout of the command
    """
    return ssh.run_command('{} {}'.format(PATH_SNAP, parameters), cwd=cwd,
                           hostname=HOSTNAME, username=USERNAME, port=PORT)


def login(email, password):
    """Login to snapd.

    :param email: Ubuntu SSO account email address.
    :param password: Ubuntu SSO account password.
    """
    ssh.run_command(COMMANDS_LOGIN.format(email=email, password=password))
    return is_logged_in(email)


def is_logged_in(email):
    """Return bool representing if the user is logged into snapd."""
    try:
        return json.loads(ssh.run_command('cat ~/.snap/auth.json')).get(
            'email') == email
    except ValueError:
        return False


def logout(email):
    """Logout snapd current user."""
    if is_logged_in(email):
        run_snapd_command_ssh(COMMAND_LOGOUT)


def download(snap, channel=CHANNEL_STABLE):
    """Download the requested snap.

    :param snap: name of the snap to download.
    :param channel: name of the release channel to download from.
    """
    command = COMMAND_DOWNLOAD.format(snap=snap, channel=channel)
    run_snapd_command_ssh(command, cwd=ssh.run_command('mktemp -d'))


def install(snap, channel=CHANNEL_STABLE):
    """Install the requested snap."""
    run_snapd_command_ssh(COMMAND_INSTALL.format(snap=snap, channel=channel))


def _is_installed(snap):
    """Return bool representing whether a snap is installed."""
    for installed_snap in _parse_output(run_snapd_command_ssh(COMMAND_LIST)):
        if installed_snap['name'] == snap:
            return True
    return False


def remove(snap):
    """Remove a snap, if its already installed."""
    if _is_installed(snap):
        run_snapd_command_ssh(COMMAND_REMOVE.format(snap=snap))


def info(snap, verbose=False):
    """Query the Ubuntu store of the information about a snap.

    :param snap: Name of the snap for which the info is required.
    :param verbose: Whether to information should be detailed.
    :return: Return a dictionary containing information about the snap.
    """
    command = COMMAND_INFO.format(snap=snap)
    if verbose:
        command = ' '.join([command, '--verbose'])
    return yaml.load("""{}""".format(run_snapd_command_ssh(command)))


def refresh(snap, channel=CHANNEL_STABLE):
    """Refresh the requested snap."""
    run_snapd_command_ssh(COMMAND_REFRESH.format(snap=snap, channel=channel))


def _parse_output(raw_output):
    """Pretty parse the output from snapd commands like `find` and `list`.

    :param raw_output: The raw output returned from the command
    :return: A list of dictionaries containing sorted results from the output.
    """
    split_output = raw_output.split('\n')
    headers = [header.lower() for header in split_output.pop(0).split()]
    return [dict(zip(headers, line.split())) for line in split_output]


def find(keyword):
    """Find snaps based on the provided filters

    :param keyword: Keyword to use for the query.
    :return: Return a list of dictionaries containing information about
        snaps matching the `keyword`.
    """
    return _parse_output(
        run_snapd_command_ssh(COMMAND_FIND.format(search_term=keyword)))
