# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Snappy Ecosystem Tests
# Copyright (C) 2017 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Module to enable staging environment for snapd."""

import logging
import sys
from time import sleep

from snappy_ecosystem_tests.environment.data.snapd import (
    CONTAINER_HOME, CONTAINER_ENV_VARS, CONTAINER_SETUP)
from snappy_ecosystem_tests.utils.lxd import Container, launch_container

LOGGER = logging.getLogger(__name__)


class StagingBuilder:
    """Class to build and activate snapd staging environment."""
    def __init__(self, container):
        if not isinstance(container, Container):
            raise RuntimeError(
                '`container` must be an instance of '
                'snappy_ecosystem_tests.utils.lxd.Container')
        self.container = container

    def _setup_staging(self):
        """Export environment variables required to enable snapd/snapcraft
        staging."""
        LOGGER.info('Exporting staging environment variables')
        updated_path = '{}/bin:{}'.format(
            CONTAINER_HOME, self.container.execute('echo $PATH'))
        CONTAINER_ENV_VARS['PATH'] = updated_path
        for key, value in CONTAINER_ENV_VARS.items():
            self.container.set_environment_variable(key, value, save=False)
        self.container.save(wait=True)
        LOGGER.info('Exported staging environment variables')

    def _wait_for_internet(self, attempts_count=10, attempt_interval=1):
        """Wait for internet connectivity in the container."""
        for _ in range(attempts_count):
            if self.container.is_internet_working():
                break
            else:
                LOGGER.info('Sleeping for %s second(s)', attempt_interval)
                sleep(attempt_interval)
        else:
            LOGGER.error(
                'Failed to get internet access in container %s after %s '
                'seconds', self.container.name,
                attempts_count * attempt_interval)
            sys.exit(1)

    def _setup_and_build(self):
        """Setup the container, pulls and builds snapd with staging enabled.
        """
        for command in CONTAINER_SETUP:
            if isinstance(command, dict):
                self.container.execute(**command)
            else:
                self.container.execute(command)

    def build(self):
        """Setup lxd container environment for snapd staging enabled builds."""
        self._setup_staging()
        self._wait_for_internet()
        self._setup_and_build()


if __name__ == '__main__':
    # TODO remove local import
    import argparse
    logging.basicConfig(level=logging.INFO)
    PARSER = argparse.ArgumentParser(
        description='Builder for snapd staging environment.')
    PARSER.add_argument('container_name', help='name for the new container.')
    PARSER.add_argument(
        'ubuntu_release', help='Ubuntu release version for the container.')
    ARGS = PARSER.parse_args()
    StagingBuilder(launch_container(name=ARGS.container_name,
                                    release=ARGS.ubuntu_release)).build()
