# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Clients needed to interact with the Snap Store"""

import contextlib
import hashlib
import itertools
import json
import logging
import os
import urllib.parse
from queue import Queue
from threading import Thread
from time import sleep

import pymacaroons
import requests
from progressbar import (
    AnimatedMarker,
    ProgressBar,
    UnknownLength,
)
from simplejson.scanner import JSONDecodeError
from snappy_ecosystem_tests.commons.config import CONFIG_STACK

from snappy_ecosystem_tests.helpers.snapcraft import constants
from snappy_ecosystem_tests.helpers.snapcraft.config import SnapcraftConfig
from snappy_ecosystem_tests.helpers.snapcraft.indicators import (
    download_requests_stream
)
from snappy_ecosystem_tests.helpers.snapcraft.options import ProjectOptions
from snappy_ecosystem_tests.helpers.store_apis import errors
from snappy_ecosystem_tests.helpers.store_apis.upload import upload_files
from snappy_ecosystem_tests.models.snap import Snap
from snappy_ecosystem_tests.utils.filters import filter_list

LOGGER = logging.getLogger(__name__)

JSON_CONTENT_TYPE = {'Content-Type': 'application/json'}
JSON_ACCEPT = {'Accept': 'application/json'}
JSON_HEADERS = dict(JSON_CONTENT_TYPE, **JSON_ACCEPT)


def _macaroon_auth(conf):
    """Format a macaroon and its associated discharge.
    :return: A string suitable to use in an Authorization header.
    """
    root_macaroon_raw = conf.get('macaroon')
    if root_macaroon_raw is None:
        raise errors.InvalidCredentialsError(
            'Root macaroon not in the config file')
    unbound_raw = conf.get('unbound_discharge')
    if unbound_raw is None:
        raise errors.InvalidCredentialsError(
            'Unbound discharge not in the config file')

    root_macaroon = _deserialize_macaroon(root_macaroon_raw)
    unbound = _deserialize_macaroon(unbound_raw)
    bound = root_macaroon.prepare_for_request(unbound)
    discharge_macaroon_raw = bound.serialize()
    auth = 'Macaroon root={}, discharge={}'.format(
        root_macaroon_raw, discharge_macaroon_raw)
    return auth


def _deserialize_macaroon(value):
    """Deserialize the given macaroon and return it
    :raise InvalidCredentialsError: when macaroon cannot be deserialized
    """
    try:
        return pymacaroons.Macaroon.deserialize(value)
    except:
        raise errors.InvalidCredentialsError('Failed to deserialize macaroon')


class Client:
    """A base class to define clients for the ols servers.
    This is a simple wrapper around requests.Session so we inherit all good
    bits while providing a simple point for tests to override when needed.
    """

    def __init__(self, conf, root_url):
        self.conf = conf
        self.root_url = root_url
        self.session = requests.Session()

    def request(self, method, url, params=None, headers=None, **kwargs):
        """Overriding base class to handle the root url."""
        # Note that url may be absolute in which case 'root_url' is ignored by
        # urljoin.
        if headers is None:
            headers = JSON_HEADERS
        final_url = urllib.parse.urljoin(self.root_url, url)
        response = self.session.request(
            method, final_url, headers=headers,
            params=params, **kwargs)
        return response

    def get(self, url, **kwargs):
        """Do a GET request and return the response"""
        return self.request('GET', url, **kwargs)

    def post(self, url, **kwargs):
        """Do a POST request and return the response"""
        return self.request('POST', url, **kwargs)

    def put(self, url, **kwargs):
        """Do a PUT request and return the response"""
        return self.request('PUT', url, **kwargs)


class Store:
    """It is an entry point for all the REST operations in the store.
    It will call the corresponding client to resolve the request.
    It also takes care about refreshing the macaroon if necessary
    in each request.
    """

    ACLs = ('package_upload', 'package_access')

    def __init__(self):
        self.conf = SnapcraftConfig()
        self.sso = SSOClient(self.conf)
        self.cpi = SnapIndexClient(self.conf)
        self.updown = UpDownClient(self.conf)
        self.sca = SCAClient(self.conf)

    def login(self, email, password, one_time_password=None, acls=ACLs,
              packages=None, channels=None, save=True):
        """Log in via the Ubuntu One SSO API."""
        # Ask the store for the needed capabilities to be associated with the
        # macaroon.
        macaroon = self.sca.get_macaroon(acls, packages, channels)
        caveat_id = self._extract_caveat_id(macaroon)
        unbound_discharge = self.sso.get_unbound_discharge(
            email, password, one_time_password, caveat_id)
        # The macaroon has been discharged, save it in the config
        self.conf.set('macaroon', macaroon)
        self.conf.set('unbound_discharge', unbound_discharge)
        if save:
            self.conf.save()

    @property
    def config(self):
        """Get current config"""
        return self.conf

    def load_config(self):
        """Load current config"""
        self.conf.load()

    def _extract_caveat_id(self, root_macaroon):
        """Extract caveat id for the given macaroon
        :raise InvalidCredentialsError: when given macaroon is not valid
        """
        macaroon = pymacaroons.Macaroon.deserialize(root_macaroon)
        # macaroons are all bytes, never strings
        sso_host = urllib.parse.urlparse(self.sso.root_url).netloc
        for caveat in macaroon.caveats:
            if caveat.location == sso_host:
                return caveat.caveat_id
        raise errors.InvalidCredentialsError('Invalid root macaroon')

    def logout(self):
        """Logout from Store client"""
        self.conf.clear()
        self.conf.save()

    def _refresh_if_necessary(self, func, *args, **kwargs):
        """Make a request, refreshing macaroons if necessary."""
        try:
            return func(*args, **kwargs)
        except errors.StoreMacaroonNeedsRefreshError:
            unbound_discharge = self.sso.refresh_unbound_discharge(
                self.conf.get('unbound_discharge'))
            self.conf.set('unbound_discharge', unbound_discharge)
            self.conf.save()
            return func(*args, **kwargs)

    def get_account_information(self):
        """Get current account information"""
        return self._refresh_if_necessary(self.sca.get_account_information)

    def get_registered_snaps(self, series=constants.DEFAULT_SERIES):
        """Get registered snaps for current user.
        First refresh macaroon if necessary"""
        return self._refresh_if_necessary(self.sca.get_registered_snaps,
                                          series)

    def filter_snaps(self, series=constants.DEFAULT_SERIES, **predicates):
        """Filter snaps that match with given predicates
        First refresh macaroon if necessary"""
        return self._refresh_if_necessary(self.sca.filter_snaps,
                                          series=series, **predicates)

    def register_key(self, account_key_request):
        """Register a key for the current user"""
        return self._refresh_if_necessary(
            self.sca.register_key, account_key_request)

    def register(self, snap_name, private=False):
        """Register a package name for the current user
        First refresh macaroon if necessary"""
        return self._refresh_if_necessary(
            self.sca.register, snap_name, private, constants.DEFAULT_SERIES)

    def push_precheck(self, snap_name):
        """Do a snap push as a pre-check (dry_run: do not actually push)
        First refresh macaroon if necessary
        """
        return self._refresh_if_necessary(
            self.sca.snap_push_precheck, snap_name)

    def push_snap_build(self, snap_id, snap_build):
        """Do a snap push
        First refresh macaroon if necessary
        """
        return self._refresh_if_necessary(
            self.sca.push_snap_build, snap_id, snap_build)

    def upload(self, snap_name, snap_filename):
        """Upload a snap file.
        :raise InvalidCredentialsError  when no unbound_discharge
         is found in current config
        """
        if not self.conf.get('unbound_discharge'):
            raise errors.InvalidCredentialsError(
                'Unbound discharge not in the config file')

        updown_data = upload_files(snap_filename, self.updown)

        return self._refresh_if_necessary(
            self.sca.snap_push_metadata, snap_name, updown_data)

    def release(self, snap_name, revision, channels):
        """Release a snap revision to the given channels
        First refresh macaroon if necessary
        """
        return self._refresh_if_necessary(
            self.sca.snap_release, snap_name, revision, channels)

    def get_snap_history(self, snap_name, series=None, arch=None):
        """Get a snap history. First refresh macaroon if necessary
        :raise: SnapNotFoundError if the snap does not belong to
        the current user
        or if it is not found.
        """
        if series is None:
            series = constants.DEFAULT_SERIES

        account_info = self.get_account_information()
        try:
            snap_id = account_info['snaps'][series][snap_name]['snap-id']
        except KeyError:
            raise errors.SnapNotFoundError(snap_name, series=series, arch=arch)

        response = self._refresh_if_necessary(
            self.sca.snap_history, snap_id, series, arch)

        if not response:
            raise errors.SnapNotFoundError(snap_name, series=series, arch=arch)

        return response

    def get_snap_status(self, snap_name, series=None, arch=None):
        """Get a snap status. First refresh macaroon if necessary
        :raise: SnapNotFoundError if the snap does not belong
        to the current user
        or if it is not found.
        """
        if series is None:
            series = constants.DEFAULT_SERIES

        account_info = self.get_account_information()
        try:
            snap_id = account_info['snaps'][series][snap_name]['snap-id']
        except KeyError:
            raise errors.SnapNotFoundError(snap_name, series=series, arch=arch)

        response = self._refresh_if_necessary(
            self.sca.snap_status, snap_id, series, arch)

        if not response:
            raise errors.SnapNotFoundError(snap_name, series=series, arch=arch)

        return response

    def close_channels(self, snap_id, channel_names):
        """Close channels for the given snap_id.
        First refresh macaroon if necessary"""
        return self._refresh_if_necessary(
            self.sca.close_channels, snap_id, channel_names)

    def download(self, snap_name, channel, download_path, arch=None):
        """Download a snap and store it locally
        :param snap_name: the snap name
        :param channel: the channel name
        :param download_path: the local path to download the snap
        :param arch: the snap architecture
        """
        if arch is None:
            arch = ProjectOptions().deb_arch

        package = self.cpi.get_package(snap_name, channel, arch)
        self._download_snap(
            snap_name, download_path,
            package['anon_download_url'], package['download_sha512'])

    def _download_snap(self, name, download_path, download_url,
                       expected_sha512):
        """Download a snap and store it locally
        :param name: the snap name
        :param download_path: the local path to download the snap
        :param download_url: the url to download the snap
        :param expected_sha512: the snap expected sha256
        """
        if self._is_downloaded(download_path, expected_sha512):
            LOGGER.info('Already downloaded %s at %s', name, download_path)
            return
        LOGGER.info('Downloading %s at %s', name, download_path)
        request = self.cpi.get(download_url, stream=True)
        request.raise_for_status()
        download_requests_stream(request, download_path)

        if self._is_downloaded(download_path, expected_sha512):
            LOGGER.info('Successfully downloaded %s at %s', name,
                        download_path)
        else:
            raise errors.SHAMismatchError(download_path, expected_sha512)

    @staticmethod
    def _is_downloaded(path, expected_sha512):
        """Return True if the snap is already downloaded in the local machine
        :param path: the path to look for the snap
        :param expected_sha512: the expected snap sha512
        """
        if not os.path.exists(path):
            return False

        file_sum = hashlib.sha512()
        with open(path, 'rb') as _f:
            for file_chunk in iter(
                    lambda: _f.read(file_sum.block_size * 128), b''):
                file_sum.update(file_chunk)
        return expected_sha512 == file_sum.hexdigest()

    def push_validation(self, snap_id, assertion):
        """Push a snap validation/assertion to the store
        :param snap_id: the snap id
        :param assertion: the snap assertion
        """
        return self.sca.push_validation(snap_id, assertion)

    def get_validations(self, snap_id):
        """Get current snap validations/assertions"""
        return self.sca.get_validations(snap_id)

    def sign_developer_agreement(self, latest_tos_accepted=False):
        """Sign developer agreement to be able to do operations in the store"""
        return self.sca.sign_developer_agreement(latest_tos_accepted)


class SSOClient(Client):
    """The Single Sign On server deals with authentication.
    It is used directly or indirectly by other servers.
    """
    def __init__(self, conf):
        super().__init__(conf, os.environ.get(
            'UBUNTU_SSO_API_ROOT_URL',
            CONFIG_STACK.config.get('production_urls', 'sso')))

    def get_unbound_discharge(self, email, password, one_time_password,
                              caveat_id):
        """Get the unbound discharge for the caveat_id.
        :param email: user email
        :param password: user password
        :param one_time_password: the one time password
        :param caveat_id: the caveat id to get the unbound discharge
        :raise StoreTwoFactorAuthenticationRequired:
        if response is unauthorized and two-factor is required
        :raise StoreAuthenticationError: if failed to get unbound discharge
        """
        data = dict(email=email, password=password,
                    caveat_id=caveat_id)
        if one_time_password:
            data['otp'] = one_time_password
        response = self.post(
            'tokens/discharge', data=json.dumps(data),
            headers=JSON_HEADERS)
        try:
            response_json = response.json()
        except JSONDecodeError:
            response_json = {}
        if response.ok:
            return response_json['discharge_macaroon']
        else:
            if (response.status_code == requests.codes.get('unauthorized') and
                    any(error.get('code') == 'twofactor-required'
                        for error in response_json.get('error_list', []))):
                raise errors.StoreTwoFactorAuthenticationRequired()
            else:
                raise errors.StoreAuthenticationError(
                    'Failed to get unbound discharge: {}'.format(
                        response.text))

    def refresh_unbound_discharge(self, unbound_discharge):
        """Refresh the given unbound discharge.
        :raise StoreAuthenticationError: if fails to refresh the
        unbound discharge.
        """
        data = {'discharge_macaroon': unbound_discharge}
        response = self.post(
            'tokens/refresh', data=json.dumps(data),
            headers=JSON_HEADERS)
        if response.ok:
            return response.json()['discharge_macaroon']
        else:
            raise errors.StoreAuthenticationError(
                'Failed to refresh unbound discharge: {}'.format(
                    response.text))


class SnapIndexClient(Client):
    """The Click Package Index knows everything about existing snaps.
    https://wiki.ubuntu.com/AppStore/Interfaces/ClickPackageIndex is the
    canonical reference.
    """
    def __init__(self, conf):
        super().__init__(conf, os.environ.get(
            'UBUNTU_STORE_SEARCH_ROOT_URL',
            CONFIG_STACK.config.get('production_urls', 'search')))

    def get_default_headers(self):
        """Return default headers for CPI requests.
        Tries to build an 'Authorization' header with local credentials
        if they are available.
        Also pin specific branded store if `SNAPCRAFT_UBUNTU_STORE`
        environment is set.
        """
        headers = {}

        with contextlib.suppress(errors.InvalidCredentialsError):
            headers['Authorization'] = _macaroon_auth(self.conf)

        branded_store = os.getenv('SNAPCRAFT_UBUNTU_STORE')
        if branded_store:
            headers['X-Ubuntu-Store'] = branded_store

        return headers

    def get_package(self, snap_name, channel, arch=None):
        """Get snap details.
        :param snap_name: the snap name
        :param channel: the snap channel
        :param arch: the snap architecture
        """
        headers = self.get_default_headers()
        headers.update({
            'Accept': 'application/hal+json',
            'X-Ubuntu-Release': constants.DEFAULT_SERIES,
        })
        if arch:
            headers['X-Ubuntu-Architecture'] = arch

        params = {
            'channel': channel,
            'fields': 'status,anon_download_url,download_url,'
                      'download_sha512,snap_id,release',
        }
        LOGGER.info('Getting details for %s', snap_name)
        url = 'api/v1/snaps/details/{}'.format(snap_name)
        resp = super().get(url, headers=headers, params=params)
        if resp.status_code != 200:
            raise errors.SnapNotFoundError(snap_name, channel, arch)
        return resp.json()


class UpDownClient(Client):
    """The Up/Down server provide upload/download snap capabilities."""

    def __init__(self, conf):
        super().__init__(conf, os.environ.get(
            'UBUNTU_STORE_UPLOAD_ROOT_URL',
            CONFIG_STACK.config.get('production_urls', 'upload')))

    def upload(self, monitor):
        """Upload a snap capability"""
        return self.post(
            urllib.parse.urljoin(self.root_url, 'unscanned-upload/'),
            data=monitor,
            headers=dict({'Content-Type': monitor.content_type},
                         **JSON_ACCEPT))


class SCAClient(Client):
    """The software center agent deals with managing snaps."""

    def __init__(self, conf):
        super().__init__(conf, os.environ.get(
            'UBUNTU_STORE_API_ROOT_URL',
            CONFIG_STACK.config.get('production_urls', 'root_api')))

    def get_macaroon(self, acls, packages=None, channels=None):
        """Get a macaroon that is used to authenticate
        request against the store
        :param acls: list of permissions to request the macaroon for.
        :param packages: list of packages to request the macaroon for.
        :param channels: list of channels to request the macaroon for.
        """
        data = {'permissions': acls}
        if packages is not None:
            data.update({'packages': packages})
        if channels is not None:
            data.update({'channels': channels})
        headers = JSON_ACCEPT
        response = self.post(
            'acl/', json=data, headers=headers)
        if response.ok:
            return response.json()['macaroon']
        else:
            raise errors.StoreAuthenticationError('Failed to get macaroon')

    @staticmethod
    def _needs_refreshed_response(response):
        """Return True if the given response needs to refresh the macaroon.
        :param response: the response to check if the
        macaroon refresh is needed.
        """
        return (
            response.status_code == requests.codes.get('unauthorized') and
            response.headers.get('WWW-Authenticate') == (
                'Macaroon needs_refresh=1'))

    def request(self, *args, **kwargs):
        """Wrapper of client request that raises an exception for
         the cases that the response macaroon needs to be refreshed
         :raise StoreMacaroonNeedsRefreshError:
         if the response needs the macaroon to be refreshed.
         """
        response = super().request(*args, **kwargs)
        if self._needs_refreshed_response(response):
            raise errors.StoreMacaroonNeedsRefreshError()
        return response

    def get_account_information(self):
        """Get the current account information
        :raise StoreAccountInformationError: if response is not ok status
        """
        auth = _macaroon_auth(self.conf)
        response = self.get(
            'account',
            headers=dict({'Authorization': auth}, **JSON_ACCEPT))
        if response.ok:
            return response.json()
        else:
            raise errors.StoreAccountInformationError(response)

    def get_registered_snaps(self, series=constants.DEFAULT_SERIES):
        """
        Get the registered snaps for the current user
        :param series: get only snaps that match with this series
        :return: the list of registered snaps
        """
        account_info = self.get_account_information()
        user_snaps = account_info.get('snaps')
        snaps_dict = user_snaps.get(series)
        snaps = []
        for name, snap_data in snaps_dict.items():
            snaps.append(Snap(name=name, **snap_data))
        return snaps

    def filter_snaps(self, series=constants.DEFAULT_SERIES, **predicates):
        """
        Return a list of the snaps that match with all the given predicates.
        :param: series: only look in the snaps that match with this series
        Examples:
        filter_snap() --> return all the snaps
        filter_snap(name='my_name') --> return the snaps that matches with
         name 'my_name'
        filter_snap(name='my_name', private=True) --> return the snaps that
         matches with name 'my_name' and are private.
        filter_snap(price='5') --> return the snaps that with price '5'.
        :return: A list of snap instances
        """
        return filter_list(self.get_registered_snaps(series=series),
                           **predicates)

    def register_key(self, account_key_request):
        """Register a key for a user with the given request
        :param account_key_request: the key to register
        :raise StoreKeyRegistrationError
        """
        data = {'account_key_request': account_key_request}
        auth = _macaroon_auth(self.conf)
        response = self.post(
            'account/account-key', data=json.dumps(data),
            headers=dict({'Authorization': auth}, **JSON_HEADERS))
        if not response.ok:
            raise errors.StoreKeyRegistrationError(response)

    def register(self, snap_name, private, series):
        """Register a snap name in the store
        :param snap_name: the name to be registered
        :param private: whether the snap name is private or not
        :param series: the snap's series to be registered
        """
        auth = _macaroon_auth(self.conf)
        data = dict(snap_name=snap_name, is_private=private,
                    series=series)
        response = self.post(
            'register-name/', data=json.dumps(data),
            headers=dict({'Authorization': auth}, **JSON_CONTENT_TYPE))
        if not response.ok:
            raise errors.StoreRegistrationError(snap_name, response)
        return True

    def snap_push_precheck(self, snap_name):
        """Do a snap push pre-check (dry_run: do not actually push)
        :raise StorePushError
        """
        data = {
            'name': snap_name,
            'dry_run': True,
        }
        auth = _macaroon_auth(self.conf)
        response = self.post(
            'snap-push/', data=json.dumps(data),
            headers=dict({'Authorization': auth}, **JSON_HEADERS))
        if not response.ok:
            raise errors.StorePushError(data['name'], response)

    def snap_push_metadata(self, snap_name, updown_data):
        """Push metadata fot the given snap
        :param snap_name: the snap name which metadata want to be updated
        :param updown_data: a dictionary that contains the metadata: upload_id,
        binary_filesize, source_uploaded
        :raise: StorePushError
        """
        data = {
            'name': snap_name,
            'series': constants.DEFAULT_SERIES,
            'updown_id': updown_data['upload_id'],
            'binary_filesize': updown_data['binary_filesize'],
            'source_uploaded': updown_data['source_uploaded'],
        }
        auth = _macaroon_auth(self.conf)
        response = self.post(
            'snap-push/', data=json.dumps(data),
            headers=dict({'Authorization': auth}, **JSON_HEADERS))
        if not response.ok:
            raise errors.StorePushError(data['name'], response)

        return StatusTracker(response.json()['status_details_url'])

    def snap_release(self, snap_name, revision, channels):
        """
        Release a snap revision to the given channel
        :param snap_name: the snap name to be released
        :param revision: the revision number to be released
        :param channels: the channels in which release the snap
        :raise: StoreReleaseError
        """
        data = {
            'name': snap_name,
            'revision': str(revision),
            'channels': channels,
        }
        auth = _macaroon_auth(self.conf)
        response = self.post(
            'snap-release/', data=json.dumps(data),
            headers=dict({'Authorization': auth}, **JSON_HEADERS))
        if not response.ok:
            raise errors.StoreReleaseError(data['name'], response)

        response_json = response.json()

        return response_json

    def push_validation(self, snap_id, assertion):
        """
        Push a validation/assertion to the store
        :param snap_id: the snap_id of the snap
        :param assertion: the validation/assertion to be pushed
        :raise: JSONDecodeError, StoreValidationError
        """
        data = {
            'assertion': assertion.decode('utf-8'),
        }
        auth = _macaroon_auth(self.conf)
        response = self.put(
            'snaps/{}/validations'.format(snap_id), data=json.dumps(data),
            headers=dict({'Authorization': auth}, **JSON_HEADERS))
        if not response.ok:
            raise errors.StoreValidationError(snap_id, response)
        try:
            response_json = response.json()
        except JSONDecodeError:
            message = ('Invalid response from the server when pushing '
                       'validations: {} {}').format(
                           response.status_code, response)
            LOGGER.debug(message)
            raise errors.StoreValidationError(
                response, message='Invalid response from the server')

        return response_json

    def get_validations(self, snap_id):
        """Get validations/assertion of a given snap
        :param snap_id: the id of the snap to get the validations for.
        :raise: JSONDecodeError, StoreValidationError
        """
        auth = _macaroon_auth(self.conf)
        response = self.get(
            'snaps/{}/validations'.format(snap_id),
            headers=dict({'Authorization': auth}, **JSON_HEADERS))
        if not response.ok:
            raise errors.StoreValidationError(snap_id, response)
        try:
            response_json = response.json()
        except JSONDecodeError:
            message = ('Invalid response from the server when getting '
                       'validations: {} {}').format(
                           response.status_code, response)
            LOGGER.debug(message)
            raise errors.StoreValidationError(
                response, message='Invalid response from the server')

        return response_json

    def push_snap_build(self, snap_id, snap_build):
        """
        Push a snap build to the store
        :param snap_id: the snap id
        :param snap_build: the snap build to push
        :raise: StoreSnapBuildError
        """
        url = 'snaps/{}/builds'.format(snap_id)
        data = json.dumps({"assertion": snap_build})

        headers = dict({'Authorization': _macaroon_auth(self.conf)},
                       **JSON_CONTENT_TYPE)
        response = self.post(url, data=data, headers=headers)
        if not response.ok:
            raise errors.StoreSnapBuildError(response)

    def snap_history(self, snap_id, series=None, arch=None):
        """
        Get a snap history
        :param snap_id: the id of the snap to get the history for.
        :param series: the series of the snap
        :param arch: the architecture of the snap
        :raise: StoreSnapHistoryError
        """
        qs = {}
        if series:
            qs['series'] = series
        if arch:
            qs['arch'] = arch
        url = 'snaps/' + snap_id + '/history'
        if qs:
            url += '?' + urllib.parse.urlencode(qs)
        auth = _macaroon_auth(self.conf)
        response = self.get(
            url,
            headers=dict({'Authorization': auth}, **JSON_HEADERS))
        if not response.ok:
            raise errors.StoreSnapHistoryError(response, snap_id, series, arch)

        response_json = response.json()

        return response_json

    def snap_status(self, snap_id, series=None, arch=None):
        """
        Get a snap status
        :param snap_id: the id of the snap to get the history for.
        :param series: the series of the snap
        :param arch: the architecture of the snap
        :raise: StoreSnapStatusError
        """
        qs = {}
        if series:
            qs['series'] = series
        if arch:
            qs['arch'] = arch
        url = 'snaps/' + snap_id + '/status'
        if qs:
            url += '?' + urllib.parse.urlencode(qs)
        auth = _macaroon_auth(self.conf)
        response = self.get(
            url,
            headers=dict({'Authorization': auth}, **JSON_HEADERS))
        if not response.ok:
            raise errors.StoreSnapStatusError(response, snap_id, series, arch)

        response_json = response.json()

        return response_json

    def close_channels(self, snap_id, channel_names):
        """
        Close channel for a given snap
        :param snap_id: the id of the snap
        :param channel_names: the channels to be closed
        :raise: StoreChannelClosingError
        """
        url = 'snaps/{}/close'.format(snap_id)
        data = {
            'channels': channel_names
        }
        headers = {
            'Authorization': _macaroon_auth(self.conf),
        }
        response = self.post(url, json=data, headers=headers)
        if not response.ok:
            raise errors.StoreChannelClosingError(response)

        try:
            results = response.json()
            return results['closed_channels'], results['channel_maps']
        except (JSONDecodeError, KeyError):
            LOGGER.debug(
                'Invalid response from the server on channel closing:\n'
                '%s %s\n%s', response.status_code, response.reason,
                response.content)
            raise errors.StoreChannelClosingError(response)

    def sign_developer_agreement(self, latest_tos_accepted=False):
        """Sign developer agreement to be able to do operations in the store"""
        auth = _macaroon_auth(self.conf)
        data = {'latest_tos_accepted': latest_tos_accepted}
        response = self.post(
            'agreement/', data=json.dumps(data),
            headers=dict({'Authorization': auth}, **JSON_HEADERS))
        if not response.ok:
            raise errors.DeveloperAgreementSignError(response)
        return response.json()


class StatusTracker:

    """Class to track the status of an operation"""

    __messages = {
        'being_processed': 'Processing...',
        'ready_to_release': 'Ready to release!',
        'need_manual_review': 'Will need manual review...',
        'processing_error': 'Error while processing...',
    }

    __error_codes = (
        'processing_error',
        'need_manual_review',
    )

    def __init__(self, status_details_url):
        self.__content = None
        self.__status_details_url = status_details_url

    def track(self):
        """Track the status"""
        queue = Queue()
        thread = Thread(target=self._update_status, args=(queue,))
        thread.start()
        widgets = ['Processing...', AnimatedMarker()]
        progress_indicator = ProgressBar(widgets=widgets, maxval=UnknownLength)
        progress_indicator.start()
        content = {}
        for indicator_count in itertools.count():
            if not queue.empty():
                content = queue.get()
                if isinstance(content, Exception):
                    raise Exception(content)
                widgets[0] = self._get_message(content)
            progress_indicator.update(indicator_count)
            if content.get('processed'):
                break
            sleep(0.1)
        progress_indicator.finish()
        self.__content = content
        return content

    def raise_for_code(self):
        """Raise an exception if content has particular error codes"""
        if any(self.__content['code'] == _k for _k in self.__error_codes):
            raise errors.StoreReviewError(self.__content)

    def _get_message(self, content):
        """Get a message from content"""
        return self.__messages.get(content['code'], content['code'])

    def _update_status(self, queue):
        """Update an operation status"""
        for content in self._get_status():
            queue.put(content)
            if content['processed']:
                break
            sleep(constants.SCAN_STATUS_POLL_DELAY)

    def _get_status(self):
        """Get an operation status"""
        connection_errors_allowed = 10
        while True:
            try:
                content = requests.get(self.__status_details_url).json()
            except (requests.ConnectionError, requests.HTTPError) as _e:
                if not connection_errors_allowed:
                    yield _e
                content = {'processed': False, 'code': 'being_processed'}
                connection_errors_allowed -= 1
            yield content
