# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Snappy Ecosystem Tests
# Copyright (C) 2017 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for Registering a snap name in the store"""

import uuid

from snappy_ecosystem_tests.helpers.snapcraft.client import Snapcraft
from snappy_ecosystem_tests.helpers.store_apis.rest_apis import Store
from snappy_ecosystem_tests.helpers.test_base import SnappyEcosystemTestCase
from snappy_ecosystem_tests.utils.storeconfig import get_store_credentials


class RegisterSnapNameTestCase(SnappyEcosystemTestCase):

    def setUp(self):
        super().setUp()
        self.store = Store()
        self.snapcraft = Snapcraft()
        self.store.login(*get_store_credentials())
        self.addCleanup(self.store.logout)
        self.snapcraft.login()
        self.addCleanup(self.snapcraft.logout)

    @staticmethod
    def _get_random_snap_name(prefix='ecosystem-test'):
        """Get a random snap name using uuid
        :param prefix: the string prefix,
        if not provided, use default one.
        """
        return '{}{}'.format(prefix, int(uuid.uuid4()))

    def test_register_public_snap_name_using_snapcraft(self):
        """Test register a public snap name via snapcraft
        and verify it via REST API"""
        # Register snap via Snapcraft
        snap_name = self._get_random_snap_name()
        self.assertTrue(self.snapcraft.register(snap_name),
                        'Unable to register snap')

        # Verify the registered snap via REST API
        snaps = self.store.filter_snaps(name=snap_name)
        self.assertEqual(1, len(snaps),
                         'Expected 1 snap with name %s but found %s'
                         % (snap_name, len(snaps)))
        snap = snaps[0]
        self.assertEqual(snap_name, snap.name)
        self.assertFalse(snap.is_private())

    def test_register_private_snap_name_using_snapcraft(self):
        """Test register a private snap name via snapcraft
        and verify it via REST API"""
        # Register snap via Snapcraft
        snap_name = self._get_random_snap_name()
        self.assertTrue(self.snapcraft.register(snap_name, private=True),
                        'Unable to register snap')

        # Verify the registered snap via REST API
        snaps = self.store.filter_snaps(name=snap_name)
        self.assertEqual(1, len(snaps),
                         'Expected 1 snap with name %s but found %s'
                         % (snap_name, len(snaps)))
        snap = snaps[0]
        self.assertEqual(snap_name, snap.name)
        self.assertTrue(snap.is_private())

    def test_register_public_snap_name_using_api(self):
        """Verify that snapcraft lists a new snap registered via
        RESTful API when using the command snapcraft list-registered"""
        # Register snap via API
        snap_name = self._get_random_snap_name()
        self.assertTrue(self.store.register(snap_name),
                        'Unable to register snap')

        # Verify the registered snap via Snapcraft
        snaps = self.snapcraft.filter_snaps(name=snap_name)
        self.assertEqual(1, len(snaps),
                         'Expected 1 snap with name %s but found %s'
                         % (snap_name, len(snaps)))
        snap = snaps[0]
        self.assertEqual(snap_name, snap.name)
        self.assertFalse(snap.is_private())

    def test_register_private_snap_name_using_api(self):
        """Verify that snapcraft lists a new private snap registered via
        RESTful API when using the command snapcraft list-registered"""
        # Register snap via API
        snap_name = self._get_random_snap_name()
        self.assertTrue(self.store.register(snap_name, private=True),
                        'Unable to register snap')

        # Verify the registered snap via Snapcraft
        snaps = self.snapcraft.filter_snaps(name=snap_name)
        self.assertEqual(1, len(snaps),
                         'Expected 1 snap with name %s but found %s'
                         % (snap_name, len(snaps)))
        snap = snaps[0]
        self.assertEqual(snap_name, snap.name)
        self.assertTrue(snap.is_private())
