# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Snappy Ecosystem Tests
# Copyright (C) 2017 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Unit tests cases for filter utils"""

from testtools import (
    TestCase,
    ExpectedException
)

from snappy_ecosystem_tests.utils.filters import (
    object_filter,
    filter_list
)


class DummyObject:  # pylint: disable=too-few-public-methods
    """Dummy object to be used in unit tests"""

    def __init__(self, **kwargs):
        for _k, _v in kwargs.items():
            setattr(self, _k, _v)


class FilterTestCase(TestCase):

    def setUp(self):
        super().setUp()
        self.obj = DummyObject(name='dummy', desc='this is dummy')
        self.obj_2 = DummyObject(name='another-dummy',
                                 desc='this is another dummy')
        self.objects = [self.obj, self.obj_2]

    def test_object_filter_with_single_argument(self):
        """Filter successfully passing one argument"""
        objects = list(filter(object_filter(name='dummy'), self.objects))
        self.assertIsNotNone(objects)
        self.assertEqual(len(objects), 1)
        self.assertEqual(objects[0], self.obj)

    def test_object_filter_with_multiple_arguments(self):
        """Filter successfully passing two arguments"""
        objects = list(filter(object_filter(name='dummy',
                                            desc='this is dummy'),
                              self.objects))
        self.assertIsNotNone(objects)
        self.assertEqual(len(objects), 1)
        self.assertEqual(objects[0], self.obj)

    def test_object_filter_does_not_match(self):
        """Return empty list when no object match criteria"""
        objects = list(filter(object_filter(name='dummy',
                                            desc='any desc'),
                              self.objects))
        self.assertEqual([], objects)

    def test_object_filter_return_all(self):
        """Get all objects when no filter is provided"""
        objects = list(filter(object_filter(), self.objects))
        self.assertEqual(len(objects), len(self.objects))

    def test_object_filter_raises_attribute_error_if_invalid_filter_name(self):
        """Object filter raises AttributeError when invalid filter
         is provided"""
        with ExpectedException(AttributeError):
            list(filter(object_filter(invalid='dummy'), self.objects))

    def test_filter_list_with_single_argument(self):
        """Filter List successfully passing one argument"""
        objects = filter_list(self.objects, name='dummy')
        self.assertIsNotNone(objects)
        self.assertEqual(len(objects), 1)
        self.assertEqual(objects[0], self.obj)

    def test_filter_list_with_multiple_arguments(self):
        """Filter List successfully passing multiple argument"""
        objects = filter_list(self.objects, name='another-dummy',
                              desc='this is another dummy')
        self.assertIsNotNone(objects)
        self.assertEqual(len(objects), 1)
        self.assertEqual(objects[0], self.obj_2)

    def test_filter_list_does_not_match(self):
        """Filter List returns empty list when no object match criteria"""
        objects = filter_list(self.objects, desc='desc does not match')
        self.assertEqual([], objects)

    def test_filter_list_return_all(self):
        """Filter list returns all objects when no filter is provided"""
        objects = filter_list(self.objects)
        self.assertEqual(len(objects), len(self.objects))

    def test_filter_list_raises_attribute_error_if_invalid_filter_name(self):
        """Filter list raises AttributeError when invalid filter is provided"""
        with ExpectedException(AttributeError):
            filter_list(self.objects, invalid='invalid filter')
