# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Snappy Ecosystem Tests
# Copyright (C) 2017 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Module to manipulate lxd containers."""

import logging
import shlex

import pylxd
from pylxd import models

DEFAULT_CONTAINER_CONFIG = {
    'name': 'snapd',
    'source': {
        'type': 'image',
        'protocol': 'simplestreams',
        'server': 'https://cloud-images.ubuntu.com/daily',
        'alias': '16.04'
    }
}
DOMAIN_PING = 'ubuntu.com'
LOGGER = logging.getLogger(__name__)


def launch_container(name, release, delete_existing=True):
    """Launch a lxd container of the requested name and release and return
    its object."""
    lxd_client = pylxd.Client()
    if delete_existing and lxd_client.containers.exists(name):
        Container(lxd_client.containers.get(name)).delete()
    container_config = DEFAULT_CONTAINER_CONFIG.copy()
    container_config['source']['alias'] = release
    container_config['name'] = name
    container = Container(
        lxd_client.containers.create(container_config, wait=True))
    container.start(wait=True)
    return container


class Container:
    """A Class representing a lxd container.

    Extends and overrides some functionalities of pylxd.models.Container
    instance through delegation since that class does not support inheritance.
    """
    def __init__(self, container_instance):
        # TODO: move to inheritance once upstream supports it.
        if not isinstance(container_instance, models.Container):
            raise RuntimeError(
                '`container_instance` must be an instance of '
                'pylxd.models.Container'
            )
        #TODO: Remove this as the StageBuilder class already has an
        # instance of the container
        self._delegate = container_instance

    def __getattr__(self, name):
        """Set the container instance as delegate."""
        return getattr(self._delegate, name)

    @staticmethod
    def _assemble_container_commands(command, cwd=''):
        """Format and split the requested command to be container friendly."""
        exec_command_base = 'sh -c \'{}\''
        if cwd:
            command = 'cd {}; {}'.format(cwd, command)
        LOGGER.info('Run command: %s', command)
        return shlex.split(exec_command_base.format(command))

    #TODO move the following methods to another place which does not require
    # a container class
    def execute(self, command, environment=None, cwd=''):
        """Execute command in the container."""
        response = self._delegate.execute(
            self._assemble_container_commands(command, cwd=cwd),
            environment=environment or {}
        )
        if response.exit_code == 0:
            LOGGER.info(response.stdout)
            return response.stdout.strip()
        raise ValueError(response.stderr)

    def set_environment_variable(self, key, value, save=True):
        """Exports environment variable to be available to
        Container.execute()."""
        self._delegate.config.update({'environment.{}'.format(key): value})
        if save:
            self._delegate.save(wait=True)

    def is_running(self):
        """Returns bool representing whether the container is running."""
        return self._delegate.status == 'Running'

    def delete(self, stop=True, wait=True):
        """Stops and deletes the container."""
        if stop and self.is_running():
            self._delegate.stop(wait=True)
        LOGGER.info('Deleting container %s', self.name)
        self._delegate.delete(wait=wait)
        LOGGER.info('Deleted container %s', self.name)

    def is_internet_working(self, ping_domain=DOMAIN_PING):
        """Return bool representing if the container have internet access."""
        try:
            self.execute('ping -c 1 {}'.format(ping_domain))
            return True
        except ValueError:
            return False
