# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Snappy Ecosystem Tests
# Copyright (C) 2017 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Module to connect to ssh client for running tests."""

import logging
import paramiko

from snappy_ecosystem_tests.utils.user import get_snapd_remote_host_credentials
from snappy_ecosystem_tests.utils.storeconfig import get_current_store
from snappy_ecosystem_tests.environment.data.snapd import STAGING_VARIABLES

LOGGER = logging.getLogger(__name__)


class SSHClient(paramiko.SSHClient):
    """Extended SSHClient."""

    def exec_command(self, command, bufsize=-1, timeout=None, get_pty=False,  # pylint: disable=arguments-differ
                     environment=None, cwd=''):
        """Execute the given command over ssh."""
        if cwd:
            command = 'cd {}; {}'.format(cwd, command)
        _, stdout, stderr = super().exec_command(command, bufsize, timeout,
                                                 get_pty, environment)
        if stdout.channel.recv_exit_status() != 0:
            raise ValueError(stderr.read().decode().strip())
        response = stdout.read().decode().strip()
        LOGGER.info(response)
        return response

    def pull(self, remote_path, local_path):
        """Pull file/directory from a remote location.

        :param remote_path: Path on the remote location.
        :param local_path: Path to download to.
        """
        sftp = self.open_sftp()
        sftp.get(remote_path, local_path)


class SSHManager:
    """Manager class to keep a pool of ssh connections."""
    # Mangle the pool so that it cannot be easily touched from outside.
    __connection_pool = []

    def __init__(self):
        raise NotImplementedError(
            'Class cannot be instantiated, use get_instance() to get a '
            'SSHClient instance.')

    @staticmethod
    def get_instance(hostname, username, port=22, **kwargs):
        """Return the instance of SSHClient from a pool of active connections,
        otherwise create a new connection and return."""
        port = int(port)
        client = SSHManager.get_client(hostname, username, port)
        if client:
            if not client.get_transport().is_active():
                client.connect(
                    hostname, username=username, port=port, **kwargs)
        else:
            client = SSHClient()
            client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            client.connect(hostname, username=username, port=port, **kwargs)
            SSHManager.__connection_pool.append(client)
        return client

    @staticmethod
    def sync():
        """Update the connection pool to only keep alive connections."""
        pool_copy = SSHManager.__connection_pool.copy()
        for client in SSHManager.__connection_pool:
            if not client.get_transport().is_alive():
                pool_copy.remove(client)
        SSHManager.__connection_pool = pool_copy

    @staticmethod
    def get_client(hostname, username, port):
        """Return the matching client from the pool based on provided
        constraints."""
        connections = [
            client for client in SSHManager.__connection_pool if
            client.get_transport().getpeername() == (hostname, port) and
            username == client.get_transport().get_username()
        ]
        return connections[0] if connections else None


def run_command(command, cwd=None, hostname=None, username=None, port=None):
    """Run the given command on remote machine over ssh.

    :param command: a string of the command to run.
    :param cwd: Current working directory for the command.
    :param hostname: The host to run command on.
    :param username: Name of the user on the remote host to login to.
    :param port: SSH port number.
    :raises ValueError: if command exits with non-zero status.
    :return: the stdout of the command.
    """
    # Take snapd remote credentials as default if they are not provided.
    # Just arbitrary... Could have been other ones.
    _hostname, _username, _port = get_snapd_remote_host_credentials()
    ssh = SSHManager.get_instance(
        hostname or _hostname, username or _username, port or _port)
    return ssh.exec_command(
        command,
        cwd=cwd,
        environment=STAGING_VARIABLES if get_current_store() == 'staging'
        else None
    )
